/* 
 * tsh - A tiny shell program with job control
 * 
 * Alok Menghrajani
 * amenghra
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>

/* Misc manifest constants */
#define MAXLINE    1024   /* max line size */
#define MAXARGS     128   /* max args on a command line */
#define MAXJOBS      16   /* max jobs at any point in time */
#define MAXJID    1<<16   /* max job ID */

/* Job states */
#define UNDEF 0 /* undefined */
#define FG 1    /* running in foreground */
#define BG 2    /* running in background */
#define ST 3    /* stopped */

/* 
 * Jobs states: FG (foreground), BG (background), ST (stopped)
 * Job state transitions and enabling actions:
 *     FG -> ST  : ctrl-z
 *     ST -> FG  : fg command
 *     ST -> BG  : bg command
 *     BG -> FG  : fg command
 * At most 1 job can be in the FG state.
 */

/* Global variables */
extern char **environ;      /* defined in libc */
char prompt[] = "tsh> ";    /* command line prompt */
int verbose = 0;            /* if true, print additional output */
int nextjid = 1;            /* next job ID to allocate */
char sbuf[MAXLINE];         /* for composing sprintf messages */

struct job_t {              /* The job struct */
    pid_t pid;              /* job PID */
    int jid;                /* job ID [1, 2, ...] */
    int state;              /* UNDEF, BG, FG, or ST */
    char cmdline[MAXLINE];  /* command line */
};
struct job_t jobs[MAXJOBS]; /* The job list */
/* End global variables */


/* Function prototypes */

/* Here are the functions that you will implement */
void eval(char *cmdline);
int builtin_cmd(char **argv);
void do_bgfg(char **argv);
void waitfg(pid_t pid);

void sigchld_handler(int sig);
void sigtstp_handler(int sig);
void sigint_handler(int sig);

/* Here are helper routines that we've provided for you */
int parseline(const char *cmdline, char **argv); 
void sigquit_handler(int sig);

void clearjob(struct job_t *job);
void initjobs(struct job_t *jobs);
int maxjid(struct job_t *jobs); 
int addjob(struct job_t *jobs, pid_t pid, int state, char *cmdline);
int deletejob(struct job_t *jobs, pid_t pid); 
pid_t fgpid(struct job_t *jobs);
struct job_t *getjobpid(struct job_t *jobs, pid_t pid);
struct job_t *getjobjid(struct job_t *jobs, int jid); 
int pid2jid(pid_t pid); 
void listjobs(struct job_t *jobs);

void usage(void);
void unix_error(char *msg);
void app_error(char *msg);
typedef void handler_t(int);
handler_t *Signal(int signum, handler_t *handler);

/*
 * main - The shell's main routine 
 */
int main(int argc, char **argv) 
{
    char c;
    char cmdline[MAXLINE];
    int emit_prompt = 1; /* emit prompt (default) */

    /* Redirect stderr to stdout (so that driver will get all output
     * on the pipe connected to stdout) */
    dup2(1, 2);

    /* Parse the command line */
    while ((c = getopt(argc, argv, "hvp")) != EOF) {
        switch (c) {
        case 'h':             /* print help message */
            usage();
	    break;
        case 'v':             /* emit additional diagnostic info */
            verbose = 1;
	    break;
        case 'p':             /* don't print a prompt */
            emit_prompt = 0;  /* handy for automatic testing */
	    break;
	default:
            usage();
	}
    }

    /* Install the signal handlers */
    Signal(SIGINT,  sigint_handler);   /* ctrl-c */
    Signal(SIGTSTP, sigtstp_handler);  /* ctrl-z */
    Signal(SIGCHLD, sigchld_handler);  /* terminated or stopped child */
    Signal(SIGQUIT, sigquit_handler);  /* so parent can cleanly terminate child*/

    /* Initialize the job list */
    initjobs(jobs);
    /* Execute the shell's read/eval loop */
    while (1) {

	/* Read command line */
	if (emit_prompt) {
	    printf("%s", prompt);                   
	    fflush(stdout);
	}
	if ((fgets(cmdline, MAXLINE, stdin) == NULL) && ferror(stdin))
	    app_error("fgets error");
	if (feof(stdin)) { /* End of file (ctrl-d) */
	    fflush(stdout);
	    exit(0);
	}

	/* Evaluate the command line */
	eval(cmdline);
	fflush(stdout);
	fflush(stderr);
    } 

    exit(0); /* control never reaches here */
}


/* Fork - wrapper for fork, which checks for errors */
int Fork() {
  pid_t pid = fork();
  if (pid < 0)
    unix_error("Fork error");
  return pid;
}

/* Sigemptyset - wrapper for sigemptyset, which checks for errors */
void Sigemptyset(sigset_t *set) {
  if (sigemptyset(set)==-1)
    unix_error("sigemptyset error");
}

/* Sigaddset - wrapper for sigaddset, which checks for errors */
void Sigaddset(sigset_t *set, int signum) {
  if (sigaddset(set, signum)==-1)
    unix_error("signaddset error");
}

/* Sigprocmask - wrapper for sigprocmask, which checks for errors */
void Sigprocmask(int how, const sigset_t *set, sigset_t *oldset) {
  if (sigprocmask(how, set, oldset)==-1)
    unix_error("sigprocmask error");
}
  
/* 
 * eval - Evaluate the command line that the user has just typed in
 * 
 * If the user has requested a built-in command (quit, jobs, bg or fg)
 * then execute it immediately. Otherwise, fork a child process and
 * run the job in the context of the child. If the job is running in
 * the foreground, wait for it to terminate and then return.  Note:
 * each child process must have a unique process group ID so that our
 * background children don't receive SIGINT (SIGTSTP) from the kernel
 * when we type ctrl-c (ctrl-z) at the keyboard.  
*/
void eval(char *cmdline) 
{
  char *argv[MAXARGS];
  int bg;
  pid_t pid;
  sigset_t mask;

  bg = parseline(cmdline, argv); /* parses the cmdline and builds the argv array. returns true if
				    it's job should be in background */
  if (!builtin_cmd(argv)) {
    /* it's not a built in command */
    /* I need to block the sigchl signal, so that
       it's impossible to kill the child before it is
       registered in the jobs list. */
    Sigemptyset(&mask);
    Sigaddset(&mask, SIGCHLD);
    Sigprocmask(SIG_BLOCK, &mask, NULL);

    /* fork is a call that returns twice. The child will enter the if loop
       but the parent will continue */
    if ((pid = Fork()) == 0) {
      /* child process must run job */

      /* unblock the sigchld signal. */
      Sigprocmask(SIG_UNBLOCK, &mask, NULL);

      /* what's the return val here ? */
      /* set the child's process group to the child's process id
	 this is important so that signals (eg ctrl-c) are not sent to all children,
	 but caught by the shell and forwarded to the right ones */
      if (setpgid(0, 0)!=0)
	unix_error("Setpgid error");

      /* use execve to execute the command. It can fail if the command is undefined */
      if (execve(argv[0], argv, environ) < 0) {
	printf("%s: Command not found.\n", argv[0]);
	/* Must use exit(0), (there was a mistake maken in class lecture, where return was instead
	   used. We must end the child process ! */
	exit(0);
      }
    }

    /* record job in joblist. Notice elegant way of passing the BG/FG parameter :) */
    if(!addjob(jobs, pid, bg ? BG : FG, cmdline))
      unix_error("addjob error");
    /* unblock the sigchld signal. */
    Sigprocmask(SIG_UNBLOCK, &mask, NULL);
    /* if the process is in foreground we need to wait for it to finish,
       otherwise if it's a background, we just print the pid/jid and return. */
    if (!bg) {
      /* wait for child to finish */
      waitfg(pid);
    } else {
      /* display pid / jid */
      printf("[%d] (%d) %s", pid2jid(pid), pid, cmdline);
      /* and continue shell.. */
    }
  }
  return;
}

/* 
 * parseline - Parse the command line and build the argv array.
 * Return true (1) if the user has requested a BG job, false 
 * if the user has requested a FG job.
 */
int parseline(const char *cmdline, char **argv) 
{
    static char array[MAXLINE]; /* holds local copy of command line */
    char *buf = array;          /* ptr that traverses command line */
    char *delim;                /* points to first space delimiter */
    int argc;                   /* number of args */
    int bg;                     /* background job? */

    strcpy(buf, cmdline);
    buf[strlen(buf)-1] = ' ';  /* replace trailing '\n' with space */
    while (*buf && (*buf == ' ')) /* ignore leading spaces */
	buf++;

    /* Build the argv list */
    argc = 0;
    while ((delim = strchr(buf, ' '))) {
	argv[argc++] = buf;
	*delim = '\0';
	buf = delim + 1;
	while (*buf && (*buf == ' ')) /* ignore spaces */
	       buf++;
    }
    argv[argc] = NULL;
    
    if (argc == 0)  /* ignore blank line */
	return 1;

    /* should the job run in the background? */
    if ((bg = (*argv[argc-1] == '&')) != 0)
	argv[--argc] = NULL;

    return bg;
}

/* 
 * builtin_cmd - If the user has typed a built-in command then execute
 *    it immediately.  
 */

/* returns 0 if not a builtin_cmd */

int builtin_cmd(char **argv) 
{
  if (argv[0]==NULL) {
    return 1;
  }
  if (strcmp(argv[0], "quit")==0) {
    /* case quit */

    /* simply end the shell */
    /* I'm not deallocating allocated memory, as kernel will handle that. */
    exit(0);
    return 1;
  } else if (strcmp(argv[0], "jobs")==0) {
    /* case jobs */
    /* simple call listjobs to list the jobs */
    listjobs(jobs);
    return 1;
  } else if (strcmp(argv[0], "bg")==0) {
    /* case bg */
    /* seperate routine to handle bg/fg since it requires some code to parse the argument */
    do_bgfg(argv);
    return 1;
  } else if (strcmp(argv[0], "fg")==0) {
    /* case fg */
    /* seperate routine to handle bg/fg since it requires some code to parse the argument */
    do_bgfg(argv);
    return 1;
  } else {
    /* default case */
    return 0;     /* not a builtin command */
  }
}

/* 
 * do_bgfg - Execute the builtin bg and fg commands
 */
void do_bgfg(char **argv) 
{
  struct job_t *job;
  /* check if we have a second argument */
  if (argv[1]==NULL) {
    /* display error if there is no second argument */
    printf("%s command requires PID or %%jobid argument\n", argv[0]);
    return;
  }
  if (argv[1][0]=='\%') {
    /* convert job to jid */
    int jid;
    /* use sscanf to remove the percent and convert string to int */
    /* not testing for match as it seems the reference implementation
       doesn't, but this isn't good programming :) */
    sscanf(argv[1], "%%%d", &jid);
    /* check if such a job exists. */
    job = getjobjid(jobs, jid);
    if (job==NULL) {
      /* display error if job doesn't exist */
      printf("%s: No such job\n", argv[1]);
      return;
    }
  } else {
    /* convert job to pid */
    int pid;
    /* this time I'm going to check for a match. (just like the refimpl) */
    if (sscanf(argv[1], "%d", &pid)==0) {
      /* display error if argument not a number */
      printf("%s: argument must be a PID or %%jobid\n", argv[0]);
      return;
    }
    job = getjobpid(jobs, pid);
    /* check to see if job exists */
    if (job==NULL) {
      /* display error if job doesn't exist */
      printf("(%d): No such process\n", pid);
      return;
    }
  }

  /* send SIGCONT to job, so that it resumes execution */
  if (kill(-job->pid, SIGCONT)<0)
    unix_error("Kill error");
  /* again, depending on fg or bg job, we must wait or continue */
  if (strcmp(argv[0], "fg")==0) {
    /* case fg */
    job->state=FG;
    waitfg(job->pid);
  } else {
    /* case bg */
    printf("[%d] (%d) %s", job->jid, job->pid, job->cmdline);
    job->state=BG;
  }
  return;
}

/* 
 * waitfg - Block until process pid is no longer the foreground process
 */
void waitfg(pid_t pid)
{
  /* here I'm not following the suggestion, as I think it's not that
     confusing to have two waitpid calls... */
     
  int status;

  /* waitpid will only return if one of the children process is
     killed (or killed since WUNTRACED is used) */
  if (waitpid(pid, &status, WUNTRACED) < 0)
    unix_error("waitfg: waitpid error");
  
  /* don't remove from list if stopped */
  if (WIFSTOPPED(status)) {
    /* change status */
    struct job_t* job = getjobpid(jobs, pid);
    job->state = ST;
    /* WSTOPSIG returns the signal that caused the child to stop (there can be many different,
       but usually it will be 20 */
    printf("Job [%d] (%d) stopped by signal %d\n", job->jid, pid, WSTOPSIG(status));
    return;
  }
  /* handle case where child was killed by a int signal */
  if (WIFSIGNALED(status)) {
    struct job_t* job = getjobpid(jobs, pid);
    /* WTERMSIG will return the signal that caused the child to die. */
    printf("Job [%d] (%d) terminated by signal %d\n", job->jid, pid, WTERMSIG(status));
  }
  /* remove from list if killed */
  if(!deletejob(jobs, pid))
    unix_error("deletejob error");
  return;
}

/*****************
 * Signal handlers
 *****************/

/* 
 * sigchld_handler - The kernel sends a SIGCHLD to the shell whenever
 *     a child job terminates (becomes a zombie), or stops because it
 *     received a SIGSTOP or SIGTSTP signal. The handler reaps all
 *     available zombie children, but doesn't wait for any other
 *     currently running children to terminate.  
 */
void sigchld_handler(int sig) 
{
  int pid;
  int status;
  /* since signals aren't queued, we need to do a waitpid here */
  /* use -1 so we can catch any children. Use WNOHANG to avoid stalling the shell */
  while ((pid = waitpid(-1, &status, WNOHANG)) != 0) {
    if (pid<0) {
      if (errno == ECHILD) {
	/* no more children to wait for */
	return;
      } else
	unix_error("waitpid error");
    }
    /* remove the job from the list. In this shell
       no info is printed. */
    if(!deletejob(jobs, pid))
      unix_error("deletejob error");
  }
  return;
}

/* 
 * sigint_handler - The kernel sends a SIGINT to the shell whenver the
 *    user types ctrl-c at the keyboard.  Catch it and send it along
 *    to the foreground job.  
 */
void sigint_handler(int sig) 
{
  /* find current foreground job */
  pid_t pid = fgpid(jobs);
  /* pid=0 means that there is no foreground job */
  if (pid!=0) {
    int jid = pid2jid(pid);
                               /* this can be improved by keep foreground_pid & jid in global
			       varibale, thus avoiding to go thourgh list (in this case
			       I'm even going through list twice */
    /* send a signal to the job so that it quits */
    printf("Job [%d] (%d) terminated by signal %d\n", jid, pid, sig);
    if(kill(-pid, sig)<0)
      unix_error("Kill error");
  }
  return;
}

/*
 * sigtstp_handler - The kernel sends a SIGTSTP to the shell whenever
 *     the user types ctrl-z at the keyboard. Catch it and suspend the
 *     foreground job by sending it a SIGTSTP.  
 */
void sigtstp_handler(int sig) 
{
  /* get current foreground job */
  pid_t pid = fgpid(jobs);
  /* pid=0 means that there is no foreground job */
  if (pid!=0) {
    /* get the job struct and modify the job->state field */
    struct job_t* job = getjobpid(jobs, pid);
    job->state = ST;
    /* send a signal so job stops. */
    printf("Job [%d] (%d) stopped by signal %d\n", job->jid, pid, sig);
    if(kill(-pid, sig)<0)
      unix_error("Kill error");
  }
  return;
}

/*********************
 * End signal handlers
 *********************/

/***********************************************
 * Helper routines that manipulate the job list
 **********************************************/

/* clearjob - Clear the entries in a job struct */
void clearjob(struct job_t *job) {
    job->pid = 0;
    job->jid = 0;
    job->state = UNDEF;
    job->cmdline[0] = '\0';
}

/* initjobs - Initialize the job list */
void initjobs(struct job_t *jobs) {
    int i;

    for (i = 0; i < MAXJOBS; i++)
	clearjob(&jobs[i]);
}

/* maxjid - Returns largest allocated job ID */
int maxjid(struct job_t *jobs) 
{
    int i, max=0;

    for (i = 0; i < MAXJOBS; i++)
	if (jobs[i].jid > max)
	    max = jobs[i].jid;
    return max;
}

/* addjob - Add a job to the job list */
int addjob(struct job_t *jobs, pid_t pid, int state, char *cmdline) 
{
    int i;
    
    if (pid < 1)
	return 0;

    for (i = 0; i < MAXJOBS; i++) {
	if (jobs[i].pid == 0) {
	    jobs[i].pid = pid;
	    jobs[i].state = state;
	    jobs[i].jid = nextjid++;
	    if (nextjid > MAXJOBS)
		nextjid = 1;
	    strcpy(jobs[i].cmdline, cmdline);
  	    if(verbose){
	        printf("Added job [%d] %d %s\n", jobs[i].jid, jobs[i].pid, jobs[i].cmdline);
            }
            return 1;
	}
    }
    printf("Tried to create too many jobs\n");
    return 0;
}

/* deletejob - Delete a job whose PID=pid from the job list */
int deletejob(struct job_t *jobs, pid_t pid) 
{
    int i;

    if (pid < 1)
	return 0;

    for (i = 0; i < MAXJOBS; i++) {
	if (jobs[i].pid == pid) {
	    clearjob(&jobs[i]);
	    nextjid = maxjid(jobs)+1;
	    return 1;
	}
    }
    return 0;
}

/* fgpid - Return PID of current foreground job, 0 if no such job */
pid_t fgpid(struct job_t *jobs) {
    int i;

    for (i = 0; i < MAXJOBS; i++)
	if (jobs[i].state == FG)
	    return jobs[i].pid;
    return 0;
}

/* getjobpid  - Find a job (by PID) on the job list */
struct job_t *getjobpid(struct job_t *jobs, pid_t pid) {
    int i;

    if (pid < 1)
	return NULL;
    for (i = 0; i < MAXJOBS; i++)
	if (jobs[i].pid == pid)
	    return &jobs[i];
    return NULL;
}

/* getjobjid  - Find a job (by JID) on the job list */
struct job_t *getjobjid(struct job_t *jobs, int jid) 
{
    int i;

    if (jid < 1)
	return NULL;
    for (i = 0; i < MAXJOBS; i++)
	if (jobs[i].jid == jid)
	    return &jobs[i];
    return NULL;
}

/* pid2jid - Map process ID to job ID */
int pid2jid(pid_t pid) 
{
    int i;

    if (pid < 1)
	return 0;
    for (i = 0; i < MAXJOBS; i++)
	if (jobs[i].pid == pid) {
            return jobs[i].jid;
        }
    return 0;
}

/* listjobs - Print the job list */
void listjobs(struct job_t *jobs) 
{
    int i;
    
    for (i = 0; i < MAXJOBS; i++) {
	if (jobs[i].pid != 0) {
	    printf("[%d] (%d) ", jobs[i].jid, jobs[i].pid);
	    switch (jobs[i].state) {
		case BG: 
		    printf("Running    ");
		    break;
		case FG: 
		    printf("Foreground ");
		    break;
		case ST: 
		    printf("Stopped    ");
		    break;
	    default:
		    printf("listjobs: Internal error: job[%d].state=%d ", 
			   i, jobs[i].state);
	    }
	    printf("%s", jobs[i].cmdline);
	}
    }
}
/******************************
 * end job list helper routines
 ******************************/


/***********************
 * Other helper routines
 ***********************/

/*
 * usage - print a help message
 */
void usage(void) 
{
    printf("Usage: shell [-hvp]\n");
    printf("   -h   print this message\n");
    printf("   -v   print additional diagnostic information\n");
    printf("   -p   do not emit a command prompt\n");
    exit(1);
}

/*
 * unix_error - unix-style error routine
 */
void unix_error(char *msg)
{
    fprintf(stderr, "%s: %s\n", msg, strerror(errno));
    exit(1);
}

/*
 * app_error - application-style error routine
 */
void app_error(char *msg)
{
    fprintf(stderr, "%s\n", msg);
    exit(1);
}

/*
 * Signal - wrapper for the sigaction function
 */
handler_t *Signal(int signum, handler_t *handler) 
{
    struct sigaction action, old_action;

    action.sa_handler = handler;  
    sigemptyset(&action.sa_mask); /* block sigs of type being handled */
    action.sa_flags = SA_RESTART; /* restart syscalls if possible */

    if (sigaction(signum, &action, &old_action) < 0)
	unix_error("Signal error");
    return (old_action.sa_handler);
}

/*
 * sigquit_handler - The driver program can gracefully terminate the
 *    child shell by sending it a SIGQUIT signal.
 */
void sigquit_handler(int sig) 
{
    if (verbose)
	printf("siquit_handler: terminating after receipt of SIGQUIT signal\n");
    exit(1);
}


